﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.IO;

namespace asprise_imaging_api
{
    /// <summary>
    /// Demo utilities.
    /// </summary>
    public class DemoUtils
    {

        public static List<DemoRecord> loadDemos()
        {
            return DemoRecord.load(parseProps(asprise_scan_scanner_twain_wia_api.Properties.Resources.demos_props));
        }

        public static Dictionary<string, string> parseProps(string content)
        {
            Dictionary<string, string> dict = new Dictionary<string, string>();
            using (StringReader reader = new StringReader(content))
            {
                string propKey = null;
                string propVal = null;
                string line = null;
                Boolean continueVal = false;
                while ((line = reader.ReadLine()) != null)
                {
                    line = line.Trim();
                    if (line.StartsWith("\\n")) // remove leading \n
                    {
                        line = line.Substring(2);
                    }

                    if (continueVal)
                    {
                        propVal += "\n" + (line.EndsWith("\\") ? line.Substring(0, line.Length - 1) : line);
                        if (line.EndsWith("\\")) // still continue
                        {
                            continue;
                        } else // done
                        {
                            dict.Add(propKey, propVal.Replace("\\\\", "\\"));
                            propKey = propVal = null;
                            continueVal = false;
                            continue;
                        }
                    }

                    if (line.Length == 0) // empty line
                    {
                        continue;
                    }

                    int posEqual = line.IndexOf('=');
                    if (posEqual > 0)
                    {
                        propKey = line.Substring(0, posEqual).Trim();
                        propVal = line.Substring(posEqual + 1,  line.Length - posEqual - 1 - (line.EndsWith("\\") ? 1 : 0)).Trim();
                        continueVal = line.EndsWith("\\");
                        if (!continueVal)
                        {
                            dict.Add(propKey, propVal.Replace("\\\\", "\\"));
                            propKey = propVal = null;
                        }
                    }
                    else
                    {
                        throw new ArgumentException("Invalid line in props: " + line);
                    }
                } // while

                // handle last line ending with \\ condition.
                if (propKey != null && propVal != null)
                {
                    dict.Add(propKey, propVal.Replace("\\\\", "\\"));
                }
            } // using

            return dict;
        }

        public static string stripComments(string s)
        {
            if (s == null)
            {
                return s;
            }

            StringBuilder sb = new StringBuilder();
            using (StringReader reader = new StringReader(s))
            {
                string line = null;
                while ((line = reader.ReadLine()) != null)
                {
                    int slashPos = -1;
                    int cursor = 0;
                    while (cursor < line.Length)
                    {
                        int currentSlashPos = line.IndexOf("//", cursor);
                        if (currentSlashPos == -1) // not found
                        {
                            slashPos = -1;
                            break;
                        }
                        else if (currentSlashPos > 0 && (line[currentSlashPos - 1] == ':' || line[currentSlashPos - 1] == '.')) // skip
                        {
                            cursor = currentSlashPos + 2;
                            continue;
                        }
                        else
                        {
                            slashPos = currentSlashPos;
                            break;
                        }
                    }

                    if (sb.Length > 0)
                    {
                        sb.Append("\n");
                    }

                    sb.Append(slashPos >= 0 ? line.Substring(0, slashPos) : line);                    
                } // which each line

                return sb.ToString();
            }
        }


        public static string getDynamicSettingsPath(string fileSimpleName)
        {
            String appData = Environment.GetFolderPath(Environment.SpecialFolder.LocalApplicationData);
            if (appData == null || appData.Length == 0)
            {
                appData = Directory.GetCurrentDirectory();
            }
            if (!appData.EndsWith("/") && !appData.EndsWith("\\"))
            {
                appData += "\\";
            }
            string path = appData + fileSimpleName;
            return path;
        }

        static string lineBreakReplacement = "[LB}";
        public static void saveDynamicSettings(Dictionary<string, string> properties, string fileSimpleName)
        {
            string path = getDynamicSettingsPath(fileSimpleName);

            System.IO.StreamWriter file = new System.IO.StreamWriter(path, false, Encoding.UTF8);

            foreach (String prop in properties.Keys.ToArray())
            {
                file.WriteLine(prop + "=" + (properties[prop] == null ? null : properties[prop].Replace("\r\n", lineBreakReplacement).Replace("\n", lineBreakReplacement)));
            }

            file.Close();
        }

        public static Dictionary<string, string> readDynamicSettings(string fileSimpleName)
        {
            string path = getDynamicSettingsPath(fileSimpleName);

            Dictionary<string, string> properties = new Dictionary<string, string>();
            if (File.Exists(path))
            {
                string[] lines = File.ReadAllLines(path, Encoding.UTF8);
                if (lines != null)
                {
                    foreach (string line in lines)
                    {
                        int equalPos = line.IndexOf('=');
                        if (equalPos <= 0)
                        {
                            continue;
                        }
                        properties.Add(line.Substring(0, equalPos).Trim(), 
                            (line.Length == equalPos + 1 ? "" : line.Substring(equalPos + 1).Trim()).Replace(lineBreakReplacement, "\n"));
                    }
                }
            }
            return properties;
        }

    } // class DemoUtils

    /// <summary>
    /// Represents a demo record.
    /// </summary>
public class DemoRecord : IComparable<DemoRecord>
{
    int order;
    String title;
    String instruction;
    String codeDisplay;
    String codeRaw;
    String iconPath;

    Boolean dialogEditable = true;
    Boolean usingDialog = true;

    public DemoRecord(String title, String instruction, String iconPath, String codeDisplay, String codeRaw, int order, Boolean usingDialog, Boolean dialogEditable) {
        this.title = title;
        this.instruction = instruction;
        this.iconPath = iconPath;
        this.codeDisplay = codeDisplay;
        this.codeRaw = codeRaw;
        this.order = order;
        this.usingDialog = usingDialog;
        this.dialogEditable = dialogEditable;
    }

    Boolean isCustom() {
        return "custom".Equals(codeRaw);
    }

    /// <summary>
    /// Title of the demo.
    /// </summary>
    /// <returns></returns>
    public String getTitle() {
        return title;
    }

    /// <summary>
    /// Instruction if any.
    /// </summary>
    /// <returns></returns>
    public String getInstruction() {
        return instruction;
    }

    /// <summary>
    /// Code to be displayed on UI.
    /// </summary>
    /// <returns></returns>
    public String getCodeDisplay() {
        return codeDisplay;
    }

    /// <summary>
    /// The actual code in JSON to be executed.
    /// </summary>
    /// <returns></returns>
    public String getCodeRaw() {
        return codeRaw;
    }

    /// <summary>
    /// Optional icon path.
    /// </summary>
    /// <returns></returns>
    public String getIconPath() {
        return iconPath;
    }

    /// <summary>
    /// Loads records from the given dictionary.
    /// </summary>
    /// <param name="props"></param>
    /// <returns></returns>
    public static List<DemoRecord> load(Dictionary<string, string> props) {
        List<DemoRecord> records = new List<DemoRecord>();
        foreach(KeyValuePair<string, string> pair in props) {
            String propName = pair.Key;
            if(propName.EndsWith("title")) {
                String title = pair.Value;
                
                String instruction = null;
                props.TryGetValue(propName.Replace("title", "instruction"), out instruction);

                String iconPath = null;
                props.TryGetValue(propName.Replace("title", "icon"), out iconPath);

                String codeDisplay = null;
                props.TryGetValue(propName.Replace("title", "code_display"), out codeDisplay);

                String codeRaw = null;
                props.TryGetValue(propName.Replace("title", "code_raw"), out codeRaw);

                string val = "false";
                Boolean dialog = props.TryGetValue((propName.Replace("title", "dialog")), out val) ? ScanUtils.equalsIgnoreCase("true", val) : false;
                Boolean editable = props.TryGetValue((propName.Replace("title", "dialog-editable")), out val) ? ScanUtils.equalsIgnoreCase("true", val) : false;

                int posFirstDash = propName.IndexOf('-');
                int posSecondDash = posFirstDash < 0 ? -1 : propName.IndexOf('-', posFirstDash + 1);
                int order = posFirstDash >= 0 && posSecondDash >= 0 ? Int32.Parse(propName.Substring(posFirstDash + 1, posSecondDash - posFirstDash - 1)) : 100;
                records.Add(new DemoRecord(title, instruction, iconPath, codeDisplay, codeRaw, order, dialog, editable));
            }
        }

        records.Sort();

        return records;
    }

    /// <summary>
    /// Compares
    /// </summary>
    /// <param name="other"></param>
    /// <returns></returns>
    public int CompareTo(DemoRecord other)
    {
        if (other == null)
        {
            return -1;
        }
        return order - other.order;
    }

    }
}
