﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.Xml.Linq;
using System.Xml.XPath;
using System.Drawing;
using System.IO;
using System.Runtime.Serialization.Json;
using System.Diagnostics;
using System.Windows.Forms;

#pragma warning disable 0168, 0649

namespace asprise_imaging_api
{
    /// <summary>
    /// Scan related utility functions.
    /// </summary>
    public class ScanUtils
    {
        /// <summary>
        /// The constant string 'true'
        /// </summary>
        public const string STR_true = "true";
        /// <summary>
        /// The constant string 'false'
        /// </summary>
        public const string STR_false = "false";

        /// <summary>
        /// Extract value as int from JSON.
        /// </summary>
        /// <param name="xe"></param>
        /// <param name="elementName"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public static int xeElementValueAsInt(XElement xe, string elementName, int defaultValue)
        {
            if (xe == null || elementName == null)
            {
                return defaultValue;
            }
            XElement element = xe.Element(elementName);
            if (element == null)
            {
                return defaultValue;
            }

            String value = element.Value;
            try
            {
                return Int32.Parse(value.Trim());
            }
            catch (Exception e)
            {
                return defaultValue;
            }
        }

        /// <summary>
        /// Extract value as decimal from JSON.
        /// </summary>
        /// <param name="xe"></param>
        /// <param name="elementName"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public static decimal xeElementValueAsDecimal(XElement xe, string elementName, decimal defaultValue)
        {
            if (xe == null || elementName == null)
            {
                return defaultValue;
            }
            XElement element = xe.Element(elementName);
            if (element == null)
            {
                return defaultValue;
            }

            String value = element.Value;
            try
            {
                return Decimal.Parse(value.Trim());
            }
            catch (Exception e)
            {
                return defaultValue;
            }
        }

        /// <summary>
        /// Extract value as string from JSON.
        /// </summary>
        /// <param name="xe"></param>
        /// <param name="elementName"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public static string xeElementValueAsString(XElement xe, string elementName, string defaultValue = null)
        {
            if (xe == null || elementName == null)
            {
                return defaultValue;
            }
            XElement element = xe.Element(elementName);
            if (element == null)
            {
                return defaultValue;
            }

            return element.Value;
        }

        /// <summary>
        /// Extract value.
        /// </summary>
        /// <param name="xe"></param>
        /// <returns></returns>
        public static object xeElementValue(XElement xe)
        {
            if (xe == null)
            {
                return null;
            }
            XAttribute xaType = xe.Attribute("type");
            if (xaType != null && xaType.Value.Equals("object")) {
                return xeElementObjectToDictionary(xe);
            } else if (xaType != null && xaType.Value.Equals("array")) {
                return xeElementArrayToList(xe);
            } else if (xaType != null && xaType.Value.Equals("number")) {
                    return Int32.Parse(xe.Value);
            } else if (canBeConvertedToNumber(xe.Value)) {
                string s = xe.Value.Trim();
                if (s.EndsWith(".0")) // remove trailing .0
                {
                    s = s.Substring(0, s.Length - 2);
                }
                if (s.Contains('.'))
                {
                    return Decimal.Parse(s);
                }
                else
                {
                    return toInteger(s);
                }
            }
            else
            {
                return xe.Value;
            }
        }

        /// <summary>
        /// Extract the whole JSON node as a dictionary.
        /// </summary>
        /// <param name="xe"></param>
        /// <returns></returns>
        public static Dictionary<string, object> xeElementObjectToDictionary(XElement xe)
        {
            if (xe == null)
            {
                return null;
            }
            Dictionary<string, object> dict = new Dictionary<string, object>();
            XAttribute xaType = xe.Attribute("type");
            if (xaType == null || !xaType.Value.Equals("object"))
            {
                throw new ImagingException("Not an object: " + xe);
            }

            IEnumerable<XElement> children = xe.Elements();
            foreach (XElement child in children)
            {
                string childName = child.Name.LocalName;
                XAttribute xaTypeChild = child.Attribute("type");
                if (xaTypeChild != null && xaTypeChild.Value.Equals("object"))
                {
                    dict.Add(childName, xeElementObjectToDictionary(child));
                }
                else if (xaTypeChild != null && xaTypeChild.Value.Equals("array"))
                {
                    dict.Add(childName, xeElementArrayToList(child));
                }
                else
                {
                    dict.Add(childName, xeElementValue(child));
                }
            }
            return dict;
        }

        /// <summary>
        /// Extract JSON array as list.
        /// </summary>
        /// <param name="xe"></param>
        /// <returns></returns>
        public static List<Object> xeElementArrayToList(XElement xe)
        {
            if (xe == null)
            {
                return null;
            }
            List<Object> list = new List<object> { };
            XAttribute xaType = xe.Attribute("type");
            if (xaType == null || !xaType.Value.Equals("array"))
            {
                throw new ImagingException("Not an array: " + xe);
            }

            IEnumerable<XElement> children = xe.Elements();
            foreach (XElement child in children)
            {
                XAttribute xaTypeChild = child.Attribute("type");
                if (xaTypeChild != null && xaTypeChild.Value.Equals("object"))
                {
                    list.Add(xeElementObjectToDictionary(child));

                }
                else if (xaTypeChild != null && xaTypeChild.Value.Equals("array"))
                {
                    list.Add(xeElementArrayToList(child));
                } else {
                    list.Add(xeElementValue(child));
                }
            }
            return list;
        }

        /// <summary>
        /// Writes int value to JSON.
        /// </summary>
        /// <param name="writer"></param>
        /// <param name="elementName"></param>
        /// <param name="value"></param>
        public static void writeElementInt(XmlDictionaryWriter writer, String elementName, int value)
        {
            writer.WriteStartElement(elementName);
            writer.WriteAttributeString("type", "number");
            writer.WriteString(value.ToString());
            writer.WriteEndElement();
        }

        /// <summary>
        /// Writes decimal value to JSON.
        /// </summary>
        /// <param name="writer"></param>
        /// <param name="elementName"></param>
        /// <param name="value"></param>
        public static void writeElementDecimal(XmlDictionaryWriter writer, String elementName, decimal value)
        {
            writer.WriteStartElement(elementName);
            writer.WriteAttributeString("type", "number");
            writer.WriteString(value.ToString());
            writer.WriteEndElement();
        }

        /// <summary>
        /// Writes dictionary value to JSON.
        /// </summary>
        /// <param name="writer"></param>
        /// <param name="elementName"></param>
        /// <param name="dict"></param>
        /// <param name="skipIfDictIsEmpty"></param>
        public static void writeElementDict(XmlDictionaryWriter writer, String elementName, Dictionary<string, string> dict, bool skipIfDictIsEmpty = true)
        {
            if (skipIfDictIsEmpty && (dict == null || dict.Count == 0))
            {
                return;
            }

            writer.WriteStartElement(elementName);
            writer.WriteAttributeString("type", "object");
            if (dict != null)
            {
                foreach(KeyValuePair<string, string> pair in dict) {
                    if (pair.Value == null)
                    {
                        writer.WriteStartElement(pair.Key);
                        writer.WriteAttributeString("type", "null");
                        writer.WriteEndElement();
                    }
                    else
                    {
                        writer.WriteElementString(pair.Key, pair.Value);
                    }
                }
            }
            writer.WriteEndElement();
        }

        /// <summary>
        /// Writes array value to JSON.
        /// </summary>
        /// <param name="writer"></param>
        /// <param name="elementName"></param>
        /// <param name="list"></param>
        /// <param name="skipIfListIsEmpty"></param>
        public static void writeElementArray(XmlDictionaryWriter writer, String elementName, List<string> list, bool skipIfListIsEmpty = true)
        {
            if (skipIfListIsEmpty && (list == null || list.Count == 0))
            {
                return;
            }

            writer.WriteStartElement(elementName);
            writer.WriteAttributeString("type", "array");
            if (list != null)
            {
                foreach (string item in list)
                {
                    if (item == null)
                    {
                        writer.WriteStartElement("item");
                        writer.WriteAttributeString("type", "null");
                        writer.WriteEndElement();
                    }
                    else
                    {
                        writer.WriteStartElement("item");
                        writer.WriteAttributeString("type", "string");
                        writer.WriteString(item);
                        writer.WriteEndElement();
                    }
                }
            }
            writer.WriteEndElement();
        }

        /// <summary>
        /// Whether the string can be converted to a number.
        /// </summary>
        /// <param name="s"></param>
        /// <returns></returns>
        public static bool canBeConvertedToNumber(string s)
        {
            decimal result;
            return decimal.TryParse(s == null ? s : s.Trim(), out result);
        }

        /// <summary>
        /// Whether the given val is of byte, short, ushort, int, uint, long or ulong.
        /// </summary>
        /// <param name="val"></param>
        /// <returns></returns>
        public static bool isInteger(Object val)
        {
            return val is sbyte || val is byte || val is short || val is ushort ||
                val is int || val is uint || val is long || val is ulong;
        }

        /// <summary>
        /// Whether the given val is of float, double or decimal.
        /// </summary>
        /// <param name="val"></param>
        /// <returns></returns>
        public static bool isFloat(Object val)
        {
            return val is float || val is double || val is decimal;
        }

        /// <summary>
        /// Whether the given val is of integer or float.
        /// </summary>
        /// <param name="val"></param>
        /// <returns></returns>
        public static bool isNumber(Object val)
        {
            return isInteger(val) || isFloat(val);
        }

        /// <summary>
        /// Any object to integer.
        /// </summary>
        /// <param name="val"></param>
        /// <param name="defaultValue">Fall back value if parse fails</param>
        /// <returns></returns>
        public static int toInteger(Object val, int defaultValue = 0)
        {
            if (val == null)
            {
                return defaultValue;
            } 
            else if (isInteger(val))
            {
                return Convert.ToInt32((int)val);
            }
            else if (val is float)
            {
                return Convert.ToInt32((float)val);
            }
            else if (val is decimal)
            {
                return Convert.ToInt32((decimal)val);
            }
            else
            {
                try
                {
                    return Convert.ToInt32(val.ToString().Trim());
                }
                catch (Exception e)
                {
                    return defaultValue;
                }
            }
        }

        /// <summary>
        /// Convert the given input to decimal or fall back to the default value.
        /// </summary>
        /// <param name="val"></param>
        /// <param name="defaultValue"></param>
        /// <returns></returns>
        public static decimal toDecimal(Object val, decimal defaultValue = 0)
        {
            if (val == null)
            {
                return defaultValue;
            }
            else if (val is decimal)
            {
                return (decimal)val;
            }
            else
            {
                try
                {
                    return Convert.ToDecimal(val.ToString().Trim());
                }
                catch (Exception e)
                {
                    return defaultValue;
                }
            }
        }

        /// <summary>
        /// Parses frame value like "(0.0, 0.0, 4.0, 6.0)"
        /// </summary>
        /// <param name="frameSpec">the value in string representation</param>
        /// <returns>4-element double array or null if parsing fails.</returns>
        public static decimal[] parseFrame(String frameSpec)
        {
            if (frameSpec == null || frameSpec.Trim().Length == 0)
            {
                return null;
            }

            // remove enclosing ()
            if (frameSpec[0] != '(')
            {
                return null;
            }
            frameSpec = frameSpec.Substring(1);

            if (frameSpec.Length == 0 || frameSpec[frameSpec.Length - 1] != ')')
            {
                return null;
            }
            frameSpec = frameSpec.Substring(0, frameSpec.Length - 1);

            String[] parts = frameSpec.Split(new char[] {',', ';'});
            if (parts == null || parts.Length < 4)
            {
                return null;
            }

            decimal[] array = new decimal[4];
            for (int i = 0; i < array.Length; i++)
            {
                array[i] = toDecimal(parts[i], -1);
            }
            return array;
        }

        /// <summary>
        /// Returns true if the two given strings equal to each other ignoring cases.
        /// </summary>
        /// <param name="s1"></param>
        /// <param name="s2"></param>
        /// <returns></returns>
        public static bool equalsIgnoreCase(String s1, String s2)
        {
            if (s1 == null && s2 == null)
            {
                return true;
            }
            if (s1 == null || s2 == null)
            {
                return false;
            }
            return s1.ToLower().Equals(s2.ToLower());
        }

        /// <summary>
        /// Normalize image format name.
        /// </summary>
        /// <param name="format"></param>
        /// <returns></returns>
        public static String normalizeImageFormat(String format)
        {
            if (format == null)
            {
                return null;
            }
            if (format.ToLower().Contains("bmp"))
            {
                return AspriseImaging.FORMAT_BMP;
            }
            else if (format.ToLower().Contains("png"))
            {
                return AspriseImaging.FORMAT_PNG;
            }
            else if (format.ToLower().Contains("jp"))
            {
                return AspriseImaging.FORMAT_JPG;
            }
            else if (format.ToLower().Contains("tif"))
            {
                return AspriseImaging.FORMAT_TIF;
            }
            else if (format.ToLower().Contains("pdf"))
            {
                return AspriseImaging.FORMAT_PDF;
            }
            return format;
        }

        /// <summary>
        /// Gets the mime type of the given image format.
        /// </summary>
        /// <param name="format"></param>
        /// <returns></returns>
        public static String getMimeType(String format)
        {
            if (format == null)
            {
                return "application/octet-stream";
            }
            if (format.ToLower().Contains("bmp"))
            {
                return "image/bmp";
            }
            else if (format.ToLower().Contains("png"))
            {
                return "image/png";
            }
            else if (format.ToLower().Contains("jp"))
            {
                return "image/jpeg";
            }
            else if (format.ToLower().Contains("tif"))
            {
                return "image/tiff";
            }
            else if (format.ToLower().Contains("pdf"))
            {
                return "application/pdf";
            }
            return "application/octet-stream";
        }

        /// <summary>
        /// Creates bitmap from base64 data.
        /// </summary>
        /// <param name="base64"></param>
        /// <returns></returns>
        public static Bitmap newBitmapFromBase64(string base64)
        {
            byte[] bytes = System.Convert.FromBase64String(base64);
            using (var ms = new MemoryStream(bytes, 0, bytes.Length))
            {
                Bitmap bitmap = new Bitmap(ms, false);
                return bitmap;
            }
        }

        /// <summary>
        /// Parses json and returns the root element.
        /// </summary>
        /// <param name="jsonRaw"></param>
        /// <returns></returns>
        public static XElement parseJson(string jsonRaw)
        {
            XmlReader reader = JsonReaderWriterFactory.CreateJsonReader(Encoding.Default.GetBytes(jsonRaw), new XmlDictionaryReaderQuotas());
            return XElement.Load(reader);
        }

        /// <summary>
        /// Convert list of objects to list of strings.
        /// </summary>
        /// <param name="objects"></param>
        /// <param name="skipNullElements"></param>
        /// <returns></returns>
        public static List<string> toListOfString(List<object> objects, bool skipNullElements = true)
        {
            List<string> list = new List<string>();
            if (objects == null)
            {
                return list;
            }
            foreach (object obj in objects)
            {
                if (obj == null && skipNullElements)
                {
                    continue;
                }
                list.Add(obj == null ? null : obj.ToString());
            }
            return list;
        }

        /// <summary>
        /// Convert dictionary to string based values.
        /// </summary>
        /// <param name="dictObjects"></param>
        /// <returns></returns>
        public static Dictionary<string, string> toDictStringString(Dictionary<string, object> dictObjects)
        {
            if(dictObjects == null) {
                return null;
            }
            Dictionary<string, string> dictStrings = new Dictionary<string, string>();
            foreach (KeyValuePair<string, object> pair in dictObjects)
            {
                dictStrings.Add(pair.Key, pair.Value == null ? null : pair.Value.ToString());
            }
            return dictStrings;
        }

        /// <summary>
        /// Preview image using windows photo viewer with fallback to default associated program.
        /// </summary>
        /// <param name="imgFile"></param>
        /// <returns></returns>
        public static bool previewImage(string imgFile)
        {
            if (imgFile == null)
            {
                return false;
            }

            if(imgFile.StartsWith(".\\")) {
                imgFile = Directory.GetCurrentDirectory() + imgFile.Substring(1);
            }
            imgFile = Environment.ExpandEnvironmentVariables(imgFile);

            Process process = null;
            if (ScanUtils.isSimpleImageFormat(imgFile))
            {
                process = Process.Start(Environment.ExpandEnvironmentVariables("%SystemRoot%\\System32\\rundll32.exe"),
                    quoteIfHaveSpaces("%ProgramFiles%\\Windows Photo Viewer\\PhotoViewer.dll") + ",ImageView_Fullscreen" + " " +
                    imgFile);
            }

            if (process == null)
            {
                process = System.Diagnostics.Process.Start(imgFile);
            }

            // Console.WriteLine("Process: " + process);
            return process != null;
        }

        /// <summary>
        /// Wraps with quotes if string contains space.
        /// </summary>
        /// <param name="s"></param>
        /// <param name="expandEnvVars"></param>
        /// <returns></returns>
        public static String quoteIfHaveSpaces(String s, bool expandEnvVars = true) {
            string processed = s != null && System.Text.RegularExpressions.Regex.IsMatch(s, @"\s") ? "\"" + s.Replace("\"", "\"\"") + "\"" : s;
            if (processed != null && expandEnvVars)
            {
                processed = Environment.ExpandEnvironmentVariables(processed);
            }
            return processed;
        }

        /// <summary>
        /// Starts the process of the given fileName optionally with args.
        /// </summary>
        /// <param name="fileName"></param>
        /// <param name="args"></param>
        /// <returns>Process if successful or null if failed.</returns>
        public static Process startProcess(String fileName, params string[] args) {
            StringBuilder sb = new StringBuilder(); // format into one liner
            foreach(string arg in args) {
                if(sb.Length > 0) {
                    sb.Append(" ");
                }
                string s = quoteIfHaveSpaces(arg);
                sb.Append(s == null ? "" : s);
            }

            Console.WriteLine("RUN: " + fileName + " " + sb.ToString());
            return Process.Start(Environment.ExpandEnvironmentVariables(fileName),
                sb.ToString());
        }

        /// <summary>
        /// Returns true if image file ends with jpg, png, bmp, tif or tiff.
        /// </summary>
        /// <param name="imgFile"></param>
        /// <returns></returns>
        public static bool isSimpleImageFormat(string imgFile)
        {
            if(imgFile == null) {
                return false;
            }
            return imgFile.ToLower().EndsWith("jpg") || imgFile.ToLower().EndsWith("png") || imgFile.ToLower().EndsWith("bmp") || 
                imgFile.ToLower().EndsWith("tif") || imgFile.ToLower().EndsWith("tiff") ;
        }

        /// <summary>
        /// Returns a newly created form with control displaying the given image.
        /// </summary>
        /// <param name="img"></param>
        /// <param name="owner"></param>
        /// <param name="zoomed"></param>
        /// <returns></returns>
        public static Form displayImageOnForm(Image img, Control owner, bool zoomed) {
            Form form = new Form();

            Screen screen = owner == null ? Screen.PrimaryScreen : Screen.FromControl(owner);
            Rectangle rectScreen = screen.WorkingArea;

            PictureBox pict = new PictureBox();
            if (zoomed)
            {
                pict.SizeMode = PictureBoxSizeMode.Zoom;
                pict.Dock = DockStyle.Fill;
                form.Controls.Add(pict);
            }
            else
            {
                pict.SizeMode = PictureBoxSizeMode.AutoSize;
                pict.Dock = DockStyle.None;

                Panel panel = new Panel();
                panel.AutoScroll = true;
                panel.AutoSize = true;
                panel.Dock = DockStyle.Fill;
                panel.Controls.Add(pict);
                form.Controls.Add(panel);
            }
            
            if (img == null)
            {
                form.Text = "Image (null)";
                form.Size = new Size(rectScreen.Width / 2, rectScreen.Height / 2);
            }
            else
            {
                form.Text = "Image " + img.Width + " x " + img.Height;
                form.Size = new Size(img.Width > rectScreen.Width ? rectScreen.Width : img.Width,
                    img.Height > rectScreen.Height ? rectScreen.Height : img.Height);
            }

            pict.Image = img;

            form.StartPosition = FormStartPosition.CenterScreen;

            return form;
        }
    }
}
